==> agent-java-0.0.3-sources.jar: dev/shortloop/agent/ShortloopApiProcessor.java <==
package dev.shortloop.agent;

import com.google.common.collect.Maps;
import dev.shortloop.agent.buffer.ApiBufferKey;
import dev.shortloop.agent.buffer.BufferManagerWorker;
import dev.shortloop.agent.buffer.DiscoveredApiBufferManager;
import dev.shortloop.agent.buffer.RegisteredApiBufferManager;
import dev.shortloop.common.models.data.APISample;

import lombok.RequiredArgsConstructor;
import org.springframework.stereotype.Component;
import org.springframework.web.util.ContentCachingRequestWrapper;
import org.springframework.web.util.ContentCachingResponseWrapper;


import javax.servlet.FilterChain;
import javax.servlet.ServletException;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import java.io.IOException;
import java.util.Collection;
import java.util.Enumeration;
import java.util.Map;

@RequiredArgsConstructor
public class ShortloopApiProcessor {
    private final RegisteredApiBufferManager registeredApiBufferManager;
    private final DiscoveredApiBufferManager discoveredApiBufferManager;

    private final SDKLogger logger;

    public void processDiscoveredApi(ShortloopSpringFilter.RequestResponseContext context, final FilterChain filterChain) throws ServletException, IOException {
        BufferManagerWorker<ApiBufferKey> worker = this.discoveredApiBufferManager.getWorker();
        if (null == worker) {
            logger.error("BufferManagerWorker is NULL inside RegisteredApiProcessor");
            doFilter(filterChain, context);
            return;
        }

        boolean canOffer = worker.canOffer(context.getApiBufferKey());
        context.setRequestPayloadOmitted(true);
        context.setResponsePayloadOmitted(true);
        try {
            doFilter(filterChain, context);
        } catch (Exception exception) {
            if (canOffer) {
                tryOffering(context, exception, worker);
            }
            throw exception;
        }
        if (canOffer) {
            tryOffering(context, null, worker);
        }
    }
    public void processRegisteredApi(ShortloopSpringFilter.RequestResponseContext context, final FilterChain filterChain) throws ServletException, IOException {
        BufferManagerWorker<ApiBufferKey> worker = this.registeredApiBufferManager.getWorker();
        if (null == worker) {
            logger.error("BufferManagerWorker is NULL inside RegisteredApiProcessor");
            doFilter(filterChain, context);
            return;
        }
        boolean canOffer = worker.canOffer(context.getApiBufferKey());
        boolean shouldCaptureRequest=false;
        boolean shouldCaptureResponse=false;
        if (canOffer) {
            shouldCaptureRequest = shouldCaptureRequest(context);
            if (shouldCaptureRequest) {
                context.setCachedRequest(wrapRequest(context.getServletRequest()));
            }
            shouldCaptureResponse = shouldCaptureResponse(context);
            if (shouldCaptureResponse) {
                context.setCachedResponse(wrapResponse(context.getServletResponse()));
            }
            context.setRequestPayloadOmitted(!shouldCaptureRequest);
            context.setResponsePayloadOmitted(!shouldCaptureResponse);
        }
        try {
            long startTime = 0;
            boolean shouldComputeLatency = canOffer && shouldCaptureRequest && shouldCaptureResponse;
            if(shouldComputeLatency) {
                startTime = System.currentTimeMillis();
            }
            doFilter(filterChain, context);
            if(shouldComputeLatency) {
                context.setLatency(System.currentTimeMillis() - startTime);
            }
        } catch (Exception exception) {
            if (canOffer) {
                tryOffering(context, exception, worker);
            }
            throw exception;
        }
        if (canOffer) {
            tryOffering(context, null, worker);
        }
    }

    private void tryOffering(ShortloopSpringFilter.RequestResponseContext context, Exception exception, BufferManagerWorker<ApiBufferKey> worker) {
        try {
            APISample apiSample = getBufferEntryForApiSample(context, exception);
            if (null != apiSample) {
                worker.offer(context.getApiBufferKey(), apiSample);
            }
        } catch (Exception e) {
        }
    }

    private void doFilter(final FilterChain filterChain, ShortloopSpringFilter.RequestResponseContext context) throws ServletException, IOException {
        final HttpServletRequest servletRequest = null != context.getCachedRequest() ? context.getCachedRequest() : context.getServletRequest();
        final HttpServletResponse servletResponse = null != context.getCachedResponse() ? context.getCachedResponse() : context.getServletResponse();
        filterChain.doFilter(servletRequest, servletResponse);
    }


    private boolean shouldCaptureRequest(ShortloopSpringFilter.RequestResponseContext context) {
        if (null != context.getApiConfig() && null != context.getApiConfig().getCaptureSample() && !context.getApiConfig().getCaptureSample()) {
            return false;
        }
        return true;
    }
    private boolean shouldCaptureResponse(ShortloopSpringFilter.RequestResponseContext context) {
        if (null != context.getApiConfig() && null != context.getApiConfig().getCaptureSample() && !context.getApiConfig().getCaptureSample()) {
            return false;
        }
        return true;
    }


    private APISample getBufferEntryForApiSample(ShortloopSpringFilter.RequestResponseContext context, Exception exception) {
        APISample apiSample = new APISample();
        apiSample.setApplicationName(context.getApplicationName());
        try {
            if (null != context.getApiConfig()) {
                apiSample.setMethod(context.getApiConfig().getMethod());
            } else {
                apiSample.setMethod(context.getObservedApi().getMethod());
            }
            apiSample.setRawUri(context.getObservedApi().getUri().getUriPath());
            apiSample.setParameters(getParameters(context.getServletRequest()));
            apiSample.setRequestHeaders(getRequestHeaders(context));
            apiSample.setResponseHeaders(getResponseHeaders(context));
            apiSample.setLatency(context.getLatency());
            apiSample.setHostName(context.getServletRequest().getServerName());
            if (null != context.getCachedRequest()) {
                apiSample.setRequestPayload(new String(context.getCachedRequest().getContentAsByteArray()));
            }
            apiSample.setStatusCode(context.getServletResponse().getStatus());
            if (null == exception) {
                if (null != context.getCachedResponse()) {
                    apiSample.setResponsePayload(new String(context.getCachedResponse().getContentAsByteArray()));
                    context.getCachedResponse().copyBodyToResponse();
                    apiSample.setStatusCode(context.getCachedResponse().getStatus());
                }
            } else {
                // this is uncaught exception, even after all exception mappers
                apiSample.setUncaughtExceptionMessage(exception.getCause().getMessage());
                apiSample.setStatusCode(500);
            }
            apiSample.setRequestPayloadOmitted(context.getRequestPayloadOmitted());
            apiSample.setResponsePayloadOmitted(context.getResponsePayloadOmitted());
        } catch (Exception e) {
            logger.error("Error create bufferEntry for API", e);
        }
        return apiSample;
    }

    private static ContentCachingRequestWrapper wrapRequest(HttpServletRequest request) {
        if (request instanceof ContentCachingRequestWrapper) {
            return (ContentCachingRequestWrapper) request;
        } else {
            return new ContentCachingRequestWrapper(request);
        }
    }

    private static ContentCachingResponseWrapper wrapResponse(HttpServletResponse response) {
        if (response instanceof ContentCachingResponseWrapper) {
            return (ContentCachingResponseWrapper) response;
        } else {
            return new ContentCachingResponseWrapper(response);
        }
    }


    private Map<String, String> getRequestHeaders(ShortloopSpringFilter.RequestResponseContext context) {
        HttpServletRequest httpServletRequest = null;
        if (null != context.getCachedRequest()) {
            httpServletRequest = context.getCachedRequest();
        } else {
            httpServletRequest = context.getServletRequest();
        }
        if (null == httpServletRequest) {
            return Maps.newHashMap();
        }
        Enumeration<String> headerNames = httpServletRequest.getHeaderNames();
        Map<String, String> headerMap = Maps.newHashMap();
        if (headerNames != null) {
            while (headerNames.hasMoreElements()) {
                String headerName = headerNames.nextElement();
                headerMap.put(headerName, httpServletRequest.getHeader(headerName));
            }
        }
        return headerMap;
    }

    private Map<String, String> getResponseHeaders(ShortloopSpringFilter.RequestResponseContext context) {
        HttpServletResponse httpServletResponse = null;
        if (null != context.getCachedRequest()) {
            httpServletResponse = context.getCachedResponse();
        } else {
            httpServletResponse = context.getServletResponse();
        }
        if (null == httpServletResponse || null == httpServletResponse.getHeaderNames()) {
            return Maps.newHashMap();
        }
        Collection<String> headerNames = httpServletResponse.getHeaderNames();
        Map<String, String> headerMap = Maps.newHashMap();
        for (String header : headerNames) {
            headerMap.put(header, httpServletResponse.getHeader(header));
        }
        headerMap.put("Content-Type", httpServletResponse.getContentType());
        return headerMap;
    }

    private Map<String, String[]> getParameters(HttpServletRequest servletRequest) {
        Map<String, String[]> servletParameterMap = servletRequest.getParameterMap();
        Map<String, String[]> parameterMap = Maps.newHashMap();
        for (String key : servletParameterMap.keySet()) {
            parameterMap.put(key, servletParameterMap.get(key));
        }
        return parameterMap;
    }
}

==> agent-java-0.0.4-sources.jar: dev/shortloop/agent/ShortloopApiProcessor.java <==
package dev.shortloop.agent;

import dev.shortloop.agent.buffer.ApiBufferKey;
import dev.shortloop.agent.buffer.BufferManagerWorker;
import dev.shortloop.agent.buffer.DiscoveredApiBufferManager;
import dev.shortloop.agent.buffer.RegisteredApiBufferManager;
import dev.shortloop.common.models.data.APISample;

import lombok.RequiredArgsConstructor;
import org.springframework.web.util.ContentCachingRequestWrapper;
import org.springframework.web.util.ContentCachingResponseWrapper;


import javax.servlet.FilterChain;
import javax.servlet.ServletException;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import java.io.IOException;
import java.util.Collection;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.Map;

@RequiredArgsConstructor
public class ShortloopApiProcessor {
    private final RegisteredApiBufferManager registeredApiBufferManager;
    private final DiscoveredApiBufferManager discoveredApiBufferManager;

    private final SDKLogger logger;

    public void processDiscoveredApi(ShortloopSpringFilter.RequestResponseContext context, final FilterChain filterChain) throws ServletException, IOException {
        BufferManagerWorker<ApiBufferKey> worker = this.discoveredApiBufferManager.getWorker();
        if (null == worker) {
            logger.error("BufferManagerWorker is NULL inside RegisteredApiProcessor");
            doFilter(filterChain, context);
            return;
        }

        boolean canOffer = worker.canOffer(context.getApiBufferKey());
        context.setRequestPayloadOmitted(true);
        context.setResponsePayloadOmitted(true);
        try {
            doFilter(filterChain, context);
        } catch (Exception exception) {
            if (canOffer) {
                tryOffering(context, exception, worker);
            }
            throw exception;
        }
        if (canOffer) {
            tryOffering(context, null, worker);
        }
    }
    public void processRegisteredApi(ShortloopSpringFilter.RequestResponseContext context, final FilterChain filterChain) throws ServletException, IOException {
        BufferManagerWorker<ApiBufferKey> worker = this.registeredApiBufferManager.getWorker();
        if (null == worker) {
            logger.error("BufferManagerWorker is NULL inside RegisteredApiProcessor");
            doFilter(filterChain, context);
            return;
        }
        boolean canOffer = worker.canOffer(context.getApiBufferKey());
        boolean shouldCaptureRequest=false;
        boolean shouldCaptureResponse=false;
        if (canOffer) {
            shouldCaptureRequest = shouldCaptureRequest(context);
            if (shouldCaptureRequest) {
                context.setCachedRequest(wrapRequest(context.getServletRequest()));
            }
            shouldCaptureResponse = shouldCaptureResponse(context);
            if (shouldCaptureResponse) {
                context.setCachedResponse(wrapResponse(context.getServletResponse()));
            }
            context.setRequestPayloadOmitted(!shouldCaptureRequest);
            context.setResponsePayloadOmitted(!shouldCaptureResponse);
        }
        try {
            long startTime = 0;
            boolean shouldComputeLatency = canOffer && shouldCaptureRequest && shouldCaptureResponse;
            if(shouldComputeLatency) {
                startTime = System.currentTimeMillis();
            }
            doFilter(filterChain, context);
            if(shouldComputeLatency) {
                context.setLatency(System.currentTimeMillis() - startTime);
            }
        } catch (Exception exception) {
            if (canOffer) {
                tryOffering(context, exception, worker);
            }
            throw exception;
        }
        if (canOffer) {
            tryOffering(context, null, worker);
        }
    }

    private void tryOffering(ShortloopSpringFilter.RequestResponseContext context, Exception exception, BufferManagerWorker<ApiBufferKey> worker) {
        try {
            APISample apiSample = getBufferEntryForApiSample(context, exception);
            if (null != apiSample) {
                worker.offer(context.getApiBufferKey(), apiSample);
            }
        } catch (Exception e) {
        }
    }

    private void doFilter(final FilterChain filterChain, ShortloopSpringFilter.RequestResponseContext context) throws ServletException, IOException {
        final HttpServletRequest servletRequest = null != context.getCachedRequest() ? context.getCachedRequest() : context.getServletRequest();
        final HttpServletResponse servletResponse = null != context.getCachedResponse() ? context.getCachedResponse() : context.getServletResponse();
        filterChain.doFilter(servletRequest, servletResponse);
    }


    private boolean shouldCaptureRequest(ShortloopSpringFilter.RequestResponseContext context) {
        if (null != context.getApiConfig() && null != context.getApiConfig().getCaptureSample() && !context.getApiConfig().getCaptureSample()) {
            return false;
        }
        return true;
    }
    private boolean shouldCaptureResponse(ShortloopSpringFilter.RequestResponseContext context) {
        if (null != context.getApiConfig() && null != context.getApiConfig().getCaptureSample() && !context.getApiConfig().getCaptureSample()) {
            return false;
        }
        return true;
    }


    private APISample getBufferEntryForApiSample(ShortloopSpringFilter.RequestResponseContext context, Exception exception) {
        APISample apiSample = new APISample();
        apiSample.setApplicationName(context.getApplicationName());
        try {
            if (null != context.getApiConfig()) {
                apiSample.setMethod(context.getApiConfig().getMethod());
            } else {
                apiSample.setMethod(context.getObservedApi().getMethod());
            }
            apiSample.setRawUri(context.getObservedApi().getUri().getUriPath());
            apiSample.setParameters(getParameters(context.getServletRequest()));
            apiSample.setRequestHeaders(getRequestHeaders(context));
            apiSample.setResponseHeaders(getResponseHeaders(context));
            apiSample.setLatency(context.getLatency());
            apiSample.setHostName(context.getServletRequest().getServerName());
            if (null != context.getCachedRequest()) {
                apiSample.setRequestPayload(new String(context.getCachedRequest().getContentAsByteArray()));
            }
            apiSample.setStatusCode(context.getServletResponse().getStatus());
            if (null == exception) {
                if (null != context.getCachedResponse()) {
                    apiSample.setResponsePayload(new String(context.getCachedResponse().getContentAsByteArray()));
                    context.getCachedResponse().copyBodyToResponse();
                    apiSample.setStatusCode(context.getCachedResponse().getStatus());
                }
            } else {
                // this is uncaught exception, even after all exception mappers
                apiSample.setUncaughtExceptionMessage(exception.getCause().getMessage());
                apiSample.setStatusCode(500);
            }
            apiSample.setRequestPayloadOmitted(context.getRequestPayloadOmitted());
            apiSample.setResponsePayloadOmitted(context.getResponsePayloadOmitted());
        } catch (Exception e) {
            logger.error("Error create bufferEntry for API", e);
        }
        return apiSample;
    }

    private static ContentCachingRequestWrapper wrapRequest(HttpServletRequest request) {
        if (request instanceof ContentCachingRequestWrapper) {
            return (ContentCachingRequestWrapper) request;
        } else {
            return new ContentCachingRequestWrapper(request);
        }
    }

    private static ContentCachingResponseWrapper wrapResponse(HttpServletResponse response) {
        if (response instanceof ContentCachingResponseWrapper) {
            return (ContentCachingResponseWrapper) response;
        } else {
            return new ContentCachingResponseWrapper(response);
        }
    }


    private Map<String, String> getRequestHeaders(ShortloopSpringFilter.RequestResponseContext context) {
        HttpServletRequest httpServletRequest = null;
        if (null != context.getCachedRequest()) {
            httpServletRequest = context.getCachedRequest();
        } else {
            httpServletRequest = context.getServletRequest();
        }
        if (null == httpServletRequest) {
            return new HashMap<>();
        }
        Enumeration<String> headerNames = httpServletRequest.getHeaderNames();
        Map<String, String> headerMap = new HashMap<>();
        if (headerNames != null) {
            while (headerNames.hasMoreElements()) {
                String headerName = headerNames.nextElement();
                headerMap.put(headerName, httpServletRequest.getHeader(headerName));
            }
        }
        return headerMap;
    }

    private Map<String, String> getResponseHeaders(ShortloopSpringFilter.RequestResponseContext context) {
        HttpServletResponse httpServletResponse = null;
        if (null != context.getCachedRequest()) {
            httpServletResponse = context.getCachedResponse();
        } else {
            httpServletResponse = context.getServletResponse();
        }
        if (null == httpServletResponse || null == httpServletResponse.getHeaderNames()) {
            return new HashMap<>();
        }
        Collection<String> headerNames = httpServletResponse.getHeaderNames();
        Map<String, String> headerMap = new HashMap<>();
        for (String header : headerNames) {
            headerMap.put(header, httpServletResponse.getHeader(header));
        }
        headerMap.put("Content-Type", httpServletResponse.getContentType());
        return headerMap;
    }

    private Map<String, String[]> getParameters(HttpServletRequest servletRequest) {
        Map<String, String[]> servletParameterMap = servletRequest.getParameterMap();
        Map<String, String[]> parameterMap = new HashMap<>();
        for (String key : servletParameterMap.keySet()) {
            parameterMap.put(key, servletParameterMap.get(key));
        }
        return parameterMap;
    }
}
